﻿// 
//
// IwGame - Cross Platform Multi-purpose Game Engine using the Marmalade SDK
//
// Developed by Matthew Hopwood of Pocketeers Limited - www.pocketeers.co.uk
//
// For updates, tutorials and more details check out my blog at www.drmop.com
//
// This code is provided free of charge and without any warranty whatsoever. The only restriction to its usage is that this header must remain intact and visible in all IwGame engine files.
// If you use this engine in your product, whilst it is not mandatory, a mention of the IwGame engine would be nice or a quick email to let us know where it is being used.
//
//

#include "s3e.h"
#include "Iw2D.h"
#include "IwGx.h"
#include "IwResManager.h"
#include "IwGxFont.h"
#include "s3eOSExec.h"

#include "IwGame.h"
#include "IwGameRender2d.h"

//
//
//
// IwGame implementation
//
//
//
void CIwGame::Init(bool enable_http)
{
	setClassType("Game");
	CurrentScene = NULL;
	NextScene = NULL;
	TouchFocus = NULL;
	PrevTouch = false;

	// Initialise Mamrlade graphics system and Iw2D module
	IwGxInit();
    Iw2DInit();
	IwGxFontInit();

	// Initialise 2D renderer
	CIwGameRender2d::Create();
	IW_GAME_RENDER2D->Init();

	// Initialise the http manager (only if requested and not already initialised)
	if (enable_http && IW_GAME_HTTP_MANAGER == NULL)
	{
		CIwGameHttpManager::Create();
		IW_GAME_HTTP_MANAGER->Init();
	}

	// Initialise the XML pool system (the XML system uses pooled memory to cut down on memory fragmentation)
	CIwGameXmlParser::PoolsInit();

	// Initialise the input system
	CIwGameInput::Create();
	IW_GAME_INPUT->Init();

	// Initialise the rsource manager
	IwResManagerInit();

	// Initialise audio system
	CIwGameAudio::Create();
	IW_GAME_AUDIO->Init();

	// Set default background colour
	SetBackgroundColour(0, 0, 0, 0);

	// Get initial time stamp
	LastFrameTime = s3eTimerGetMs();

	// Initialise the global resource system
	CIwGameGlobalResources::Create();
	IW_GAME_GLOBAL_RESOURCES->Init();

	// Init the XOML system
	CIwGameXoml::Create();
	IW_GAME_XOML->Init();
}

void CIwGame::SetBackgroundColour(uint8 r, uint8 g, uint8 b, uint8 a)
{
	// Set the default background clear colour
	IwGxSetColClear(r, g, b, a);
}

void CIwGame::Release()
{
	// Clean up the XOML system
	IW_GAME_XOML->Release();
	CIwGameXoml::Destroy();

	// Clean up scenes
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
		delete *it;
	Scenes.clear();
	SceneRemovals.clear();

	// Shut down global resource system
	IW_GAME_GLOBAL_RESOURCES->Release();
	CIwGameGlobalResources::Destroy();

	// Shut down the input system
	IW_GAME_INPUT->Release();
	CIwGameInput::Destroy();

	// Shut down the resource manager
	IwResManagerTerminate();

	// Shut down audio
	IW_GAME_AUDIO->Release();
	CIwGameAudio::Destroy();

	// Destroy XML pools
	CIwGameXmlParser::PoolsDestroy();

	// Shut down http manager
	if (IW_GAME_HTTP_MANAGER != NULL)
	{
		IW_GAME_HTTP_MANAGER->Release();
		CIwGameHttpManager::Destroy();
	}

	// Shut down 2D renderer
	IW_GAME_RENDER2D->Release();
	CIwGameRender2d::Destroy();

	// Shut down Marmalade graphics system and the Iw2D module
	IwGxFontTerminate();
	Iw2DTerminate();
	IwGxTerminate();
}

bool CIwGame::Update()
{
	// Calculate how long the last game frame took (capped at 60 and 10 fps) - We use this to scale all of our trasient variables that rely upon time so that everything moves at the same rate
	// regardless of our devices frame rate
	float dt = (float)(s3eTimerGetMs() - LastFrameTime) / FRAME_SPEED_LOCK_MS;
	if (dt < 1.0) dt = 1.0f;
	if (dt > 6.0f) dt = 6.0f;
	LastFrameTime = s3eTimerGetMs();

	// Clear the screen
	IwGxClear(IW_GX_COLOUR_BUFFER_F | IW_GX_DEPTH_BUFFER_F);

	// Update input system
	IW_GAME_INPUT->Update();

	// Update audio
	IW_GAME_AUDIO->Update();

	// Check for scene change
	// Notify scenes that there is a change of circumstances
	if (CurrentScene != NextScene)
	{
		if (CurrentScene != NULL)
		{
			CurrentScene->NotifyLostFocus(NextScene);
			if (CurrentScene->getAllowSuspend())
				CurrentScene->NotifySuspending(NextScene);
		}
		if (NextScene != NULL)
		{
			NextScene->NotifyGainedFocus(CurrentScene);
			if (NextScene->getAllowSuspend())
				NextScene->NotifyResuming(CurrentScene);
		}
		CurrentScene = NextScene;
	}

	// Update all scenes that are not suspended
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if (CurrentScene == *it)
		{
			(*it)->Update(dt);
		}
		else
		{
			if (!(*it)->getAllowSuspend())
			{
				(*it)->Update(dt);
			}
		}
	}

	// Remove any deleted scenes
	for (_Iterator it = SceneRemovals.begin(); it != SceneRemovals.end(); ++it)
	{
		for (_Iterator it2 = Scenes.begin(); it2 != Scenes.end(); ++it2)
		{
			if (*it2 == *it)
			{
				if (CurrentScene == *it2)
					CurrentScene = NULL;
				if (NextScene == *it2)
					NextScene = NULL;
				delete *it2;
				Scenes.erase(it2);
				break;
			}
		}
	}
	SceneRemovals.clear();

	// Handle tapped event
	ProcessEvents(CurrentScene);

	// Process focus events for any other scenes that are marked as AllowFocus
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if (CurrentScene != *it && (*it)->getAllowFocus())
			ProcessEvents(*it);
	}

	// Update http manager
	if (IW_GAME_HTTP_MANAGER != NULL)
	{
		IW_GAME_HTTP_MANAGER->Update();
	}

	CIwGameAnimTimeline::UpdateRunCount();

	return true;
}

void CIwGame::ProcessEvents(CIwGameScene *scene)
{
	if (scene == NULL)
		return;

	bool touch = IW_GAME_INPUT->isTouching();
//	CIwFVec2 pos = CurrentScene->ScreenToVirtual((float)IW_GAME_INPUT->getTouch(0)->x, (float)IW_GAME_INPUT->getTouch(0)->y);
	CIwVec2 pos(IW_GAME_INPUT->getTouchedPos().x, IW_GAME_INPUT->getTouchedPos().y);

/*	bool tapped = IW_GAME_INPUT->hasTapped();
	if (tapped)
	{
		CIwGameActor* tapped_actor = scene->FindTappedActor(pos.x, pos.y);
		if (tapped_actor != NULL)
		{
			tapped_actor->OnTapped(pos.x, pos.y);
		}
	}*/

	// Handle begin touch event
	if (!scene->getPrevTouch() && touch)
	{
		CIwGameActor* tapped_actor = scene->FindTappedActor(pos.x, pos.y);
		if (tapped_actor != NULL)
		{
			tapped_actor->OnBeginTouch(pos.x, pos.y);
			TouchFocus = tapped_actor;
		}
	}

	// Handle end touch event (only actor that received begin touch event needs to handle this)
	if (scene->getPrevTouch() && !touch)
	{
		if (TouchFocus != NULL)
		{
			TouchFocus->OnEndTouch(pos.x, pos.y);
			if (TouchFocus->HitTest(pos.x, pos.y))
			{
				TouchFocus->OnTapped(pos.x, pos.y);
			}
		}
		TouchFocus = NULL;
	}

	scene->setPrevTouch(touch);
}

void CIwGame::Draw()
{
	// Draw all scenes that are not suspended
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		(*it)->Draw();
	}

	// Take xcare of any post draw rendering
	PostDraw();

	// Flush IwGx
	IwGxFlush();

	// Display the rendered frame
	IwGxSwapBuffers();

	// Show the surface
//	Iw2DSurfaceShow();

	// Yield to the operating system
	s3eDeviceYield(0);
}

void CIwGame::PostDraw()
{
}

void CIwGame::addScene(CIwGameScene *scene, bool bring_to_front)
{
	Scenes.push_back(scene);
	scene->setParent(this);

	// If another scene has focus then bring it back to the top of the scene stack
	if (CurrentScene != NULL && !bring_to_front)
	{
		BringSceneToFront(CurrentScene);
	}
}

void CIwGame::removeScene(CIwGameScene* scene)
{
	// Mark scene for removal
	 if (!scene->isDestroyed())
	 {
		SceneRemovals.push_back(scene);
		scene->setDestroyed(true);
	 }
}

void CIwGame::removeScene(unsigned int name_hash)
{
	// Find the scene
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if ((*it)->getNameHash() == name_hash && !(*it)->isDestroyed())
		{
			// Mark scene for removal
			SceneRemovals.push_back(*it);
			(*it)->setDestroyed(true);
			break;
		}
	}
}

CIwGameScene* CIwGame::findScene(unsigned int name_hash)
{
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if ((*it)->getNameHash() == name_hash && !(*it)->isDestroyed())
		{
			return *it;
		}
	}

	return NULL;
}

CIwGameScene* CIwGame::findScene(const char* name)
{
	return findScene(IwHashString(name));
}

CIwGameScene* CIwGame::findScene(int type)
{
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
	{
		if ((*it)->getType() == type)
		{
			return *it;
		}
	}

	return NULL;
}

CIwGameScene* CIwGame::getScene(int index)
{
	if (index < 0 || index >= Scenes.size())
		return NULL;

	return Scenes.element_at(index);
}

void CIwGame::clearScenes()
{
	for (_Iterator it = Scenes.begin(); it != Scenes.end(); ++it)
		delete *it;

	Scenes.clear();
}

void CIwGame::changeScene(CIwGameScene *new_scene)
{
	NextScene = new_scene;
}

bool CIwGame::changeScene(unsigned int name_hash)
{
	NextScene = findScene(name_hash);
	if (NextScene == NULL)
		return false;

	return true;
}

void CIwGame::BringSceneToFront(CIwGameScene* scene)
{
	for (_Iterator it = begin(); it != end(); it++)
	{
		if (*it == scene)
		{
			Scenes.erase(it);
			Scenes.push_back(scene);
			break;
		}
	}
}

bool CIwGame::LoadFromXoml(IIwGameXomlResource* parebt, bool load_children, CIwGameXmlNode* node)
{
	// Process inner tags
	IW_GAME_XOML->Process(this, node);

	return true;
}

//
//
//
//
// CIwGameCreator implementation
//
//
//
//
IIwGameXomlResource* CIwGameCreator::CreateInstance(IIwGameXomlResource* parent)
{
	return NULL;
}

//
//
//
//
//  CIwGameXomlAction_Global implementation
//
//
//
//
void CIwGameXomlAction_Global::Execute(IIwGameXomlResource* source, CIwGameAction* action)
{
	CIwGame* game = NULL;
	CIwGameScene* scene = NULL;
	CIwGameActor* actor = NULL;

	if (source->getClassTypeHash() == CIwGameXomlNames::Scene_Hash)
	{
		scene = (CIwGameScene*)source;
		game = scene->getParent();
	}
	else
	if (source->getClassTypeHash() == CIwGameXomlNames::Actor_Hash)
	{
		actor = (CIwGameActor*)source;
		scene = actor->getScene();
		game = scene->getParent();
	}

	switch (Type)
	{
	case Action_PlayTimeline:
	case Action_StopTimeline:
	case Action_SetTimeline:
		{
			CIwGameAnimTimeline* timeline = NULL;
			if (!action->Parameter2.IsEmpty())
			{
				// If user provided second paremater then take that as a target actor or scene
				IIwGameXomlResource* other = scene->findActor(action->Parameter2.getHash());
				if (other == NULL)
				{
					actor = NULL;
					scene = game->findScene(action->Parameter2.getHash());
				}
				else
				{
					actor = (CIwGameActor*)other;
					scene = actor->getScene();
				}
			}
			if (!action->Parameter1.IsEmpty())
				timeline = (CIwGameAnimTimeline*)scene->getResourceManager()->findResource(action->Parameter1.getHash(), CIwGameXomlNames::Timeline_Hash);

			if (actor != NULL)
			{
				if (timeline == NULL)
				{
					if (Type == Action_SetTimeline)
					{
						actor->setTimeline(NULL);
						return;
					}
					else
						timeline = actor->getTimeline();
				}
				if (timeline == NULL)
				{
					CIwGameError::LogError("Warning: Actions - Actor timeline was not found - ", action->Parameter1.c_str());
					return;
				}
				if (Type == Action_SetTimeline)
				{
					actor->setTimeline(timeline);
					timeline->restart();
				}
			}
			else
			{
				if (timeline == NULL)
				{
					if (Type == Action_SetTimeline)
					{
						scene->setTimeline(NULL);
						return;
					}
					else
						timeline = scene->getTimeline();
				}
				if (timeline == NULL)
				{
					CIwGameError::LogError("Warning: Actions - Scene timeline was not found - ", action->Parameter1.c_str());
					return;
				}
				if (Type == Action_SetTimeline)
				{
					scene->setTimeline(timeline);
					timeline->restart();
				}
			}
			if (Type == Action_PlayTimeline)
			{
				timeline->restart();
			}
			else
			if (Type == Action_StopTimeline)
				timeline->stop();
		}
		break;
	case Action_PlaySound:
		IW_GAME_AUDIO->PlaySound(action->Parameter1.getHash());
		break;
	case Action_PlayMusic:
		if (action->Parameter2.IsEmpty())
			IW_GAME_AUDIO->PlayMusic(action->Parameter1.c_str(), 1);
		else
			IW_GAME_AUDIO->PlayMusic(action->Parameter1.c_str(), action->Parameter2.GetAsInt());
		break;
	case Action_StopMusic:
		IW_GAME_AUDIO->StopMusic();
		break;
	case Action_LoadXOML:
		{
			IIwGameXomlResource* parent = NULL;
			if (!action->Parameter1.IsEmpty())
				parent = scene->getResourceManager()->findResource(action->Parameter2.getHash(), CIwGameXomlNames::Scene_Hash);
			if (parent == NULL)
				parent = game;
			IW_GAME_XOML->Process(parent, action->Parameter1.c_str());
		}
		break;
	case Action_SetVariable:
		{
			CIwGameXomlVariable* var = NULL;
			// First check the sceen to see if the variable is available
			if (scene != NULL)
			{
				var = scene->getVariableManager()->findVariable(action->Parameter1.c_str());
				if (var != NULL)
					var->setValue(action->Parameter2.c_str());
			}
			// Variable was not available in the scene so check the global resources
			if (var == NULL)
			{
				var = IW_GAME_GLOBAL_RESOURCES->getVariableManager()->findVariable(action->Parameter1.c_str());
				if (var != NULL)
					var->setValue(action->Parameter2.c_str());
			}
		}
		break;
	case Action_Launch:
		{
			s3eOSExecExecute(action->Parameter1.c_str(), false);
		}
		break;
	case Action_SetBGColour:
		{
			int components[8];
			if (action->Parameter1.GetAsListOfInt(components) == 4)
				game->SetBackgroundColour((uint8)components[0], (uint8)components[1], (uint8)components[2], (uint8)components[3]);
			else
				CIwGameError::LogError("Warning: Actions - SetBGColour - Incorrect component count - ", action->Parameter2.c_str());
		}
		break;
	case Action_SetCurrentScene:
		{
			CIwGameScene* scene = game->findScene(action->Parameter1.getHash());
			if (scene != NULL)
			{
				game->changeScene(scene);
				game->BringSceneToFront(scene);
			}
			else
				CIwGameError::LogError("Warning: Actions - SetCurrentScene - scene not found - ", action->Parameter1.c_str());
		}
		break;
	case Action_BringtSceneToFront:
		{
			CIwGameScene* scene = game->findScene(action->Parameter1.getHash());
			if (scene != NULL)
			{
				game->BringSceneToFront(scene);
			}
			else
				CIwGameError::LogError("Warning: Actions - BringtSceneToFront - scene not found - ", action->Parameter1.c_str());
		}
		break;
	}
}
